import UUID from 'uuid/v4'
import SparkMD5 from 'spark-md5'
import EXIF from 'exif-js'
import EE from 'event-emitter'
import FileSaver from 'file-saver'
if(!window.$emit){
  window.$emitter = EE()
}

export default class FileAgent {
  CHUNK_SIZE = 1024*1024
  UPDATE_INTERVAL = 1000
  DB = {}
  token = ''
  apiRoot = ''
  
  
  constructor(){
    window.$emitter.on('UPLOAD_FILE_CANCEL', ({fileUploadId})=>{
      this._update({progress: 'cancel'}, fileUploadId)
    })
    window.$emitter.on('RE_UPLOAD_FILE', ({fileUploadId})=>{
      let message = this.DB['fileUpload'][fileUploadId]
      let file = this.DB[`file_${fileUploadId}`]
      this._update({
        percent: message.file.start/file.size,
        progress: 'uploading',
      }, fileUploadId)
      this._reUpload(
        message.channel,
        file,
        message.file.name,
        message.file.fileUploadId,
        message.file.fileMd5,
        message.file.start
      )
    })
    window.$emitter.on('UPLOAD_FILE_GET', ({channelId})=>{
      let fileInfos = this.DB['fileUpload'] || {}
      let messages = []
      Object.keys(fileInfos).forEach(fileUploadId=>{
        let message = fileInfos[fileUploadId]
        if(Date.now()-message.ts > 24*60*60*1000){
          delete this.DB[`file_${fileUploadId}`]
          delete fileInfos[fileUploadId]
        }else if(message.channel === channelId){
          messages.push(message)
        }
      })
      this.DB['fileUpload'] = fileInfos
      if(messages.length > 0){
        window.$emitter.emit(
          'UPLOAD_FILE_ADD_BATCH',
          {messages}
        )
      }
    })
    window.$file = window.$file || this
  }

  _toast(
    msg = "",
    level = 'error',
  ){
    window.$emitter.emit('TOAST', {msg, level})
  }

  _naming(file, fileUploadId){
    let hashName = ''
    if (fileUploadId) {
      hashName = fileUploadId.replace(/(-)/g, '').substring(0, 8);
    }
    if (file.name) {
      return encodeURIComponent(file.name);
    } else if (['jpeg', 'gif', 'png', 'jpg', 'bmp'].includes(file.type.split('/')[1])) {
      if (hashName) {
        return `${hashName}.png`;
      }
      return `${Date.now()}.png`;
    }
    if (hashName) {
      return `${hashName}`;
    }
    return `${Date.now()}`;
  }

  _hash(arrayBuffer){
    let spark = new SparkMD5()
    let chars = new Uint8Array(arrayBuffer)
    const CHUNK_SIZE = 0x8000
    let index = 0
    let length = chars.length
    let result = ''
    let slice = null
    while (index < length) {
      slice = chars.subarray(index, Math.min(index + CHUNK_SIZE, length))
      result += String.fromCodePoint.apply(null, slice)
      index += CHUNK_SIZE
    }
    spark.appendBinary(result)
    return spark.end()
  }

  _base64ArrayBuffer(arrayBuffer){
    let base64 = '';
    const encodings = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';

    const bytes = new Uint8Array(arrayBuffer);
    const byteLength = bytes.byteLength;
    const byteRemainder = byteLength % 3;
    const mainLength = byteLength - byteRemainder;

    let a;
    let b;
    let c;
    let d;
    let chunk;

    // Main loop deals with bytes in chunks of 3
    for (let i = 0; i < mainLength; i += 3) {
      // Combine the three bytes into a single integer
      chunk = (bytes[i] << 16) | (bytes[i + 1] << 8) | bytes[i + 2];

      // Use bitmasks to extract 6-bit segments from the triplet
      a = (chunk & 16515072) >> 18; // 16515072 = (2^6 - 1) << 18
      b = (chunk & 258048) >> 12; // 258048   = (2^6 - 1) << 12
      c = (chunk & 4032) >> 6; // 4032     = (2^6 - 1) << 6
      d = chunk & 63;        // 63       = 2^6 - 1

      // Convert the raw binary segments to the appropriate ASCII encoding
      base64 += encodings[a] + encodings[b] + encodings[c] + encodings[d];
    }

    // Deal with the remaining bytes and padding
    if (byteRemainder === 1) {
      chunk = bytes[mainLength];

      a = (chunk & 252) >> 2; // 252 = (2^6 - 1) << 2

      // Set the 4 least significant bits to zero
      b = (chunk & 3) << 4; // 3   = 2^2 - 1

      base64 += `${encodings[a]}${encodings[b]}==`;
    } else if (byteRemainder === 2) {
      chunk = (bytes[mainLength] << 8) | bytes[mainLength + 1];

      a = (chunk & 64512) >> 10; // 64512 = (2^6 - 1) << 10
      b = (chunk & 1008) >> 4; // 1008  = (2^6 - 1) << 4

      // Set the 2 least significant bits to zero
      c = (chunk & 15) << 2; // 15    = 2^4 - 1

      base64 += `${encodings[a]}${encodings[b]}${encodings[c]}=`;
    }

    return base64;
  }

  _dealWithImage(
    arrayBuffer, 
    type, 
    fileUploadId
  ){
    if(!['jpeg', 'gif', 'png', 'jpg', 'bmp'].includes(type.split('/')[1]))  return

    let data = this._base64ArrayBuffer(arrayBuffer)
    let img = new Image()
    img.onload = ()=>{
      let imageStyle = {
        width: img.width>320 ? 320 : img.width,
        height: img.width>320 ? img.height*320/img.width : img.height
      }
      EXIF.getData(img, function(){
        if(EXIF.getTag(this, 'Orientation') === 3){
          imageStyle.transform = `rotate(180deg) translate(${imageStyle.width}, ${imageStyle.height})`
        }else if(EXIF.getTag(this, 'Orientation') === 6){
          imageStyle.transform = `rotate(90deg) translate(${imageStyle.width}, ${imageStyle.height})`
        }else if(EXIF.getTag(this, 'Orientation') === 8){
          imageStyle.transform = `rotate(-90deg) translate(${imageStyle.width}, ${imageStyle.height})`
        }
      })
      this._update(
        {
          thumbnailUrl: `data:${type};base64,${data}`,
          imageStyle,
          progress: 'uploading',
        }, 
        fileUploadId
      )
    }
    img.src = `data:${type};base64,${data}`
  }

  _uploadFragment(
    channelId,
    file,
    fileName,
    fileUploadId,
    start = 0,
    status = 0,
  ){
    let end = start+this.CHUNK_SIZE>file.size ? file.size : start+this.CHUNK_SIZE
    let uploadBlob = file.slice(start, end)
    this._upload(
      `/file/upload/${channelId}`,
      uploadBlob,
      {
        version: 'v1.0.1',
        params: {
          fileName,
          fileSize: file.size,
          fileUploadId,
        },
        progress: percent=>{
          percent = start/file.size + percent*(end-start)/file.size
          this._update({
            percent,
            progress: 'uploading',
          }, fileUploadId)
        },
      },
      status,
    ).then(res=>{
      if (res.resultCode === '000000'){
        if(res.resultDetail.uploadFlag === 'true'){
          this._update(
            {progress: status===1 ? 'cancel' : 'success'}, 
            fileUploadId
          )
        }else if(res.resultDetail.uploadFlag === 'false'){
          if(status!==1 && end<file.size){
            if(!this.DB['fileUpload'] || !this.DB['fileUpload'][fileUploadId]){
              this._uploadFragment(
                channelId, 
                file, 
                fileName,
                fileUploadId, 
                end,
                1
              )
            }else{
              this._uploadFragment(
                channelId, 
                file, 
                fileName,
                fileUploadId, 
                end,
                status,
              )
            }
          }
        }
      }else if(res.resultCode === '100102'){
        this._toast('不可上传0字节文件')
        this._update({
          progress: 'failure',
          start
        }, fileUploadId)
      }else{
        this._toast('文件发送失败')
        this._update({
          progress: 'failure',
          start
        }, fileUploadId)
      }
    })
  }

  _upload(
    url,
    file,
    {
      version = 'v1.0.0',
      params = {},
      progress = ()=>{},
    }={},
    uploadStatus = 0,
  ){
    let queryString = Object.keys(params).reduce((s, k)=>{
      s += `&${k}=${encodeURIComponent(params[k])}`
      return s
    },  "")
    if(queryString){
      queryString = url.includes('?') ? queryString : ("?"+queryString.substring(1))
    }
    const FILE_SIZE = file.size
    return new Promise((resolve, reject)=>{
      let xhr = new XMLHttpRequest()
      xhr.open('POST', `${this.apiRoot}/${version}${url}${queryString}`)
      xhr.setRequestHeader('Content-Type', 'application/octet-stream;charset=UTF-8')
      xhr.setRequestHeader('Authorization', `bearer ${thistoken}`)
      xhr.setRequestHeader('uploadStatus', uploadStatus)
      xhr.upload.addEventListener('loadstart', e=>{
        progress(e.loaded/FILE_SIZE)
      })
      xhr.upload.addEventListener('progress', e=>{
        progress(e.loaded/FILE_SIZE)
      })
      xhr.addEventListener('error', e=>{
        resolve({resultCode: '999999'})
      })
      xhr.addEventListener('loadend', e=>{
        if(xhr.status === 200){
          resolve(JSON.parse(xhr.responseText))
        }else if(xhr.status === 401){
          resolve({resultCode: '888888'})
        }
      })
      xhr.send(file)
    })
  }

  _save(
    channelId,
    file,
    fileName,
    fileUploadId,
    user,
    lastMessage,
    fileMd5,
  ){
    let fileType = file.type
    if (['jpeg', 'gif', 'png', 'jpg', 'bmp'].includes(fileType.split('/')[1])) {
      fileType = 'uploadImg/uploadImg';
    }
    if(file){
      this.DB[`file_${fileUploadId}`] = file
    }

    let message = {
      id: fileUploadId,
      channel: channelId,
      file: {
        fileUploadId,
        fileType: fileType ? fileType.split('/')[1] : 'rar',
        mimeType: fileType || 'application/x-rar-compressed',
        name: fileName,
        size: file.size,
        fileMd5,
        start: 0,
        percent: 0,
        progress: 'start',
        upLoadType: 'file', // 上传文件 分为  file  img 两种类型展示
      },
      user: user,
      type: 'EVENT_MESSAGE',
      subtype: 'FILE_SHARE',
      ts: lastMessage.ts+1,
    }

    window.$emitter.emit(
      'UPLOAD_FILE_ADD',
      {message}
    )

    const fileInfos = this.DB['fileUpload'] || {}
    this.DB['fileUpload'] = {
      ...fileInfos,
      [fileUploadId]: {
        ...message
      }
    }
  }

  _update(
    fileInfo,
    fileUploadId,
  ){
    const fileInfos = this.DB['fileUpload'] || {}
    if(!fileInfos[fileUploadId]) return

    if(fileInfo.progress==='cancel'
      || fileInfo.progress==='success'
      || fileInfo.progress==='exist'){
      window.$emitter.emit(
        'UPLOAD_FILE_REMOVE',
        {
          messageId: fileUploadId
        }
      )
      delete this.DB[`up_prev_${fileUploadId}`]
      delete this.DB[`file_${fileUploadId}`]
      delete fileInfos[fileUploadId]
      this.DB['fileUpload'] = fileInfos
    }else{
      fileInfos[fileUploadId].file = {
        ...fileInfos[fileUploadId].file,
        ...fileInfo,
      }
      this.DB['fileUpload'] = fileInfos

      let prev = this.DB[`up_prev_${fileUploadId}`] || 0
      let cur = Date.now()
      let isUpdate = true
      if(this.DB[`file_${fileUploadId}`].size > 10*1024*1024){
        isUpdate = cur-prev > this.UPDATE_INTERVAL
      }else{
        isUpdate = cur-prev > this.UPDATE_INTERVAL/2
      }
      if(fileInfo.progress==='uploading' && fileInfo.percent!==0 &&fileInfo.percent!==1 && !isUpdate){
        return
      }else{
        this.DB[`up_prev_${fileUploadId}`] = cur
      }
      if(fileInfo.progress==='failure'){
        delete this.DB[`up_prev_${fileUploadId}`]
      }
      window.$emitter.emit(
        'UPLOAD_FILE_UPDATE',
        {
          message: fileInfos[fileUploadId]
        }
      )
    }
  }

  upload(
    channelId,
    files=[],
    user,
    lastMessage,
  ){
    if(files.length === undefined){
      files = [files]
    }
    [...files].forEach(file=>{
      let fileUploadId = UUID()
      let fileName = this._naming(file, fileUploadId)
      if(fileName.length > 300){
        this._toast('文件名过长发送失败')
      }else if(file.size >= 100*1024*1024){
        this._toast('文件大小超过100M！发送失败')
      }else{
        let fileReader = new FileReader()
        fileReader.onload = e=>{
          let fileMd5 = this._hash(e.target.result)
          HttpAgent.File.check(
            channelId,
            fileName,
            fileUploadId,
            fileMd5,
          ).then(res=>{
            if(res.resultCode === '000000'){
              if(res.resultDetail === 'false'){
                // nothing
              }else{
                this._save(
                  channelId,
                  file,
                  fileName,
                  fileUploadId,
                  user,
                  lastMessage,
                  fileMd5,
                )
                this._dealWithImage(
                  e.target.result,
                  file.type,
                  fileUploadId,
                )
                this._uploadFragment(
                  channelId,
                  file,
                  fileName,
                  fileUploadId,
                )
              }
            }else{
              this._toast(res.resultMessage)
            }
          })
        }
        fileReader.readAsArrayBuffer(file)
      }
    })
  }

  _reUpload(
    channelId,
    file,
    fileName,
    fileUploadId,
    fileMd5,
    start,
  ){
    HttpAgent.File.check(
      channelId,
      fileName,
      fileUploadId,
      fileMd5,
    ).then(res=>{
      if(res.resultCode === '000000'){
        if(res.resultDetail === 'false'){
          this._update({progress: 'exist'}, fileUploadId)
        }else{
          this._uploadFragment(
            channelId,
            file,
            fileName,
            fileUploadId,
            start,
            2
          )
        }
      }else{
        this._toast(res.resultMessage)
      }
    })
  }

  download(
    fileId,
    channelId,
  ){
    let prev = 0
    let xhr = new XMLHttpRequest()
    xhr.open('GET', `${this.apiRoot}/v1.0.0/file/download/${fileId}/${channelId}`)
    xhr.setRequestHeader("Content-type", "application/json")
    xhr.setRequestHeader('Authorization', `bearer ${this.token}`)
    xhr.responseType = 'blob'
    xhr.addEventListener('loadstart', e=>{
      window.$emitter.emit('DOWNLOAD_FILE',{
        fileId,
        percent: 0,
        progress: 'start',
      })
    })
    xhr.addEventListener('progress', e=>{
      let cur = Date.now()
      let isUpdate = true
      if(e.total > 10*1024*1024){
        isUpdate = cur-prev > this.UPDATE_INTERVAL
      }else{
        isUpdate = cur-prev > this.UPDATE_INTERVAL/2
      }
      if(e.loaded!==e.total && !isUpdate){
        return
      }else{
        prev = cur
      }
      window.$emitter.emit('DOWNLOAD_FILE',{
        fileId,
        percent: e.loaded/e.total,
        progress: 'downloading',
        cancel: ()=>{xhr.abort()}
      })
    })
    xhr.addEventListener('abort', e=>{
      window.$emitter.emit('DOWNLOAD_FILE', {
        fileId,
        progress: 'cancel'
      })
    })
    xhr.addEventListener('error', e=>{
      this._toast('文件下载失败')
      window.$emitter.emit('DOWNLOAD_FILE', {
        fileId,
        progress: 'failure'
      })
    })
    xhr.addEventListener('loadend', e=>{
      prev = null
      if(xhr.status === 200){
        window.$emitter.emit('DOWNLOAD_FILE',{
          fileId,
          percent: e.loaded/e.total,
          progress: 'end',
        })
        FileSaver.saveAs(xhr.response, decodeURIComponent(xhr.getResponseHeader('filename')))
      }else if(xhr.status === 401){
        this._toast('认证失败')
        window.$emitter.emit('DOWNLOAD_FILE', {
          fileId,
          progress: 'failure'
        })
      }
    })
    xhr.send()
  }
}

